import numpy as np
import numpy.linalg as la
import numpy.random as random
import sys,os,pdb
#from mpi4py import MPI
import time
"""
Finds the compound nucleus (CN) resonances for a model of neutron-nucleus scattering.
Based on the Mazama code of G. F. Bertsch.
Adapted by P. Fanto
Last updated: 8/14/2017
Dependencies: numpy, mpi4py
"""

# Make the neutron Hamiltonian on a spatial mesh
def mk_Hn(V0,a,r0,A,delr,Nd,h2over2M):
	R = r0*A**(1./3.)
	i_entry = int(R/delr)-1
	t = h2over2M/(delr**2)
	H = np.zeros((Nd,Nd))
	V = np.zeros(Nd)
	for i in range(Nd):
		rn = (i+1)*delr
		V[i] = V0/(1.0 + np.exp((rn-R)/a))
		H[i,i] = 2*t + V[i]
		if i > 0:
		    H[i,i-1] = H[i-1,i] = -t
	return H,V,t, i_entry

# Make the diagonal CN states.  Spacings follow the middle third of a random-matrix spectrum.
def mk_Hc(Nc,E0,dE,gamb):
  Hb = np.array([[0.0j]*Nc]*Nc)
  eb = np.zeros(Nc)
  E1 = E0 + dE*Nc
  eb = wigner_mod(E0,E1,Nc)
  for i in range(Nc):
    Hb[i,i] = eb[i] - 1.0j*gamb/2.
  return Hb

# Create a spectrum of N energies between E1 and E2 that follows the middle third of a
# GOE spectrum.
def wigner_mod(E1,E2,N):
  #if N > 200:
    #print 'N too large'
    #sys.exit()
  m1=np.random.randn(3*N,3*N)

  m3=m1+np.transpose(m1)
  e,u = la.eigh(m3)
  e.sort()
  # select middle third
  emin = e[N]
  emax = e[2*N]

  erange = emax-emin
  scale = (E2-E1)/erange
  eout = np.array([0.0]*N)
  for i in range(N):
     eout[i] = (e[i+N]-emin)*scale + E1
  return eout


start_time = time.time()

h2over2M = 20.75 #hbar^2/2m

# input file
lines = open(sys.argv[1]).readlines()
print lines[0]

# read in parameters for Woods-Saxon potential and spatial mesh
V0, a_r, r0, A = map(float,lines[1].split()[:4])
sp = lines[2].split()
delr = float(sp[0]); Nd = int(sp[1])

print 'V0, a_r, r0, A = ', V0,a_r,r0,A
print 'delr, number of spatial mesh sites = ',delr,Nd

# Make neutron Hamiltonian
Hn,V,tr,i_entry = mk_Hn(V0,a_r,r0,A,delr,Nd,h2over2M)

# read in seed for the random number generator and number of GOE realizations
seed, Ntrials = map(int,lines[3].split()[:2])
#random.seed(seed)
print 'seed, number of realizations = ', seed,Ntrials
# read in parameters for CN Hamiltonian
sp = lines[4].split()
Nint,Nbuf = map(int,sp[:2])
Nc = Nint + 2*Nbuf
EI, dE, Ggamma = map(float,sp[2:5])
print 'Nc,Nbuf,lowest internal energy above buffer (keV), level spacing (eV), \
	highest internal state energy (keV), gamma width (meV) = ',Nc,EI*1000,dE*1e6,1000*(EI + dE*Nc),Ggamma*1e9
Ntot = Nc + Nd
Elow = EI - dE*Nbuf # lowest energy in buffer

# read in v values
sp = lines[5].split()
vvals = np.array([])
for i in range(len(sp)):
	if sp[i] == '#':
		break
	vvals = np.append(vvals,float(sp[i]))

# Initialize full Hamiltonian
H = np.zeros((Ntot,Ntot))*1j
H[0:Nd,0:Nd] = Hn
iden = np.identity(Ntot)

# Matrices required to find resonances
P = h2over2M*iden
C = np.zeros((Ntot,Ntot)); C[Nd-1,Nd-1] = 1.

# Iteration parameters
tol = 1e-12 # numerical tolerance for root finding
Nsteps = 1000 # maximum number of iterative steps taken

# read in output file prefix
prefix = str(lines[6].split()[0])

# Find resonances for all coupling constants
#pdb.set_trace()
for indv in range(len(vvals)):
	random.seed(seed) # seed the random number generator
	
	v = vvals[indv]
	print 'v = ',v

	trialsKept = 0 # number of successful trials for this value of v

	kres = np.array([])*1j # array of resonances k's
	
	# loop over realizations of the GOE
	print 'realizations of GOE'
	for trial in range(Ntrials):
		trialstart = time.time()
		# make CN Hamiltonian and add to full Hamiltonian
		#Hb = mk_Hc(Nc,EI,dE,Ggamma)
		Hb = mk_Hc(Nc,Elow,dE,Ggamma)
		Ec = np.diag(Hb)
		Nguess = np.array([0])
		H[Nd:Ntot,Nd:Ntot] = Hb
		'''
		kguess = np.zeros(Nc)*1j
		for i in range(Nc):
			kg = np.sqrt(Ec[i])
			if kg.imag > 0: 
				kg *= -1.0
			kguess[i] = kg
		'''
		
		# maximum energy for CN resonances, used as a "gate"
		Eguess_max = 2*max(Ec.real) 
		Eguess_min = -1.5*min(Ec.real)

		# make coupling matrix and add to full Hamiltonian
		vb = np.zeros((Nd,Nc))*1j
		tmp = random.randn(Nc)
		vb[i_entry,:] = v*tmp
		H[0:Nd,Nd:Ntot] = vb
		H[Nd:Ntot,0:Nd] = vb.T

		# Find initial guesses via quadratic approximation to full matrix M(k)

		# Construct U = H - tr*C
		U = H*(1. + 0.0j); U[Nd-1,Nd-1] -= tr

		# Construct V
		V = np.zeros((Ntot,Ntot))*1j; V[Nd-1,Nd-1] = 1j*delr*tr
		# V = 1j*delr*tr*C

		# Construct W.   W = h2over2M*I - .5*tr*delr**2 C
		W = h2over2M*iden; W[Nd-1,Nd-1] -= .5*tr*delr**2
		# Construct W^-1
		Winv = (1./h2over2M)*iden
		Winv[Nd-1,Nd-1] = 1./(h2over2M-.5*delr**2*tr)

		# Construct and diagonalize matrix for initial guesses
		BinvA = np.zeros((2*Ntot,2*Ntot))*1j
		BinvA[0:Ntot,Ntot:2*Ntot] = iden
		BinvA[Ntot:2*Ntot,0:Ntot] = np.dot(Winv,U)
		BinvA[Ntot:2*Ntot,Ntot:2*Ntot] = np.dot(Winv,-V)
		kvals = la.eig(BinvA)[0]

		#pdb.set_trace()
		Nk = np.array([len(kvals)])

		
		# Use "gates" to get initial guesses for compound resonances
		# Conditions: Re(E) > 0 and Re(E) < 2*maximum internal energy
		kguess = np.array([])*1j
		Egmin = Ec[Nbuf-2].real; Egmax = Ec[Nbuf + Nint-1].real
		#print Egmin,Egmax
		index = 0
		for i in range(Nk):
			kt = kvals[i]
			Ek = h2over2M*kt**2
			if Ek.real >= Egmin and Ek.real <= Egmax:
				if -2*Ek.imag >= Ggamma:
					kguess = np.append(kguess,kt)
					index += 1
		if len(kguess) < Nint:
			print 'Trial failed because gating eliminated too many guesses'
			continue
		step = 0
		while len(kguess) > Nint:
			if step % 2 == 0:
				kguess = kguess[1:]
			else:
				kguess = kguess[:-1]
			step += 1
		Nguess = np.array([len(kguess)])

		print 'Nguess for trial', trial, '= ', Nguess
		#for i in range(Nguess):
		#	print kguess[i]
		#sys.exit()
		#continue
		kconv = np.array([])*1j # number of converged resonances for this trial

		# iterate over guesses to find actual roots
		for i in range(Nguess):
			#loop_time = time.time()

			kt = kguess[i]
			E = h2over2M*kt**2

			# Construct Mk
			Mk = H*(1.0+0.0j)
			for j in range(Ntot):
				Mk[j,j] -= E
			Mk[Nd-1,Nd-1] -= tr*np.exp(1j*kt*delr)
			lk = la.eig(Mk)[0]
			lmin = min(abs(lk))

			# if already a root, converge
			if lmin <= tol:
				# assert resonance selection criteria
				Et = h2over2M*kt**2
				gn = -2*Et.imag - Ggamma
				if Et.real > 0 and gn > 0:
					kconv = np.append(kconv,kt)
				
			# otherwise, iterate to find the root.  Maximum number of iterations is 1000.
			else:
				kg = kt*1.0
				for step in range(Nsteps):
					Mguess = H*(1. + 0.0j) - kg**2*P - np.exp(1j*kg*delr)*tr*C
					# check for convergence
					if step > 1:
						lmin = min(abs(la.eig(Mguess)[0]))
						if lmin <= tol:
							# resonance selection criteria
							Eg = h2over2M*kg**2
							gn = -2*Eg.imag - Ggamma
							if Eg.real > 0 and gn > 0:
								kconv = np.append(kconv,kg)
							break
					# iteration based on generalized eigenvalue problem
					Mprime_inv = (1./(-2.*kg*h2over2M))*iden
					Mprime_inv[Nd-1,Nd-1] = (1./(-2*kg*h2over2M - 1j*delr*tr*np.exp(1j*kg*delr)))
					S = np.dot(Mprime_inv,Mguess)
					ls = la.eig(S)[0]
					istep = np.argmin(abs(ls))
					lstep = ls[istep]
					kg -= lstep
			#print time.time()-loop_time
	
		# if Nint resonances aren't found, eliminate the trial
		if len(kconv) < Nint:
			print 'trial ', trial, 'failed because too few resonances were found'
		elif len(kconv) == Nint:
			print 'trial ',trial, ' successful'
			kres = np.append(kres,kconv)
			trialsKept += 1
		else:
			Econv = h2over2M*kconv**2
			sortind = np.argsort(Econv)
			sortind = sortind[:Nint]
			kres = np.append(kres,kconv[sortind])
			print 'trial', trial,' successful'
			trialsKept += 1
		print 'trial time = ', time.time()-trialstart, ' sec'
		#print 'resonances found during trial ', trial+1, ' = ', count
		

	# output results
	Nres = len(kres)
	print 'Nres from ',Ntrials,' trials is ',Nres
	Eres = h2over2M*kres**2 # resonance energies

	# centroid and neutron width of resonances
	E0 = Eres.real
	Gn = -2*Eres.imag - Ggamma

	
	#print time,' seconds'
	of = prefix + '-Ntrials' + str(Ntrials) + '-v%1.1e-V%2.2f' % (v,abs(V0))
	of += '.dat'
	with open(of,'w') as fl:
		fl.write('# Statistical results, internal resonances \n')
		fl.write('# Number of successful trials = ' + str(trialsKept) + '\n')
		fl.write('# delr,Nd,V0 = ' + str(delr) + ' ' + str(Nd) + ' ' + str(V0) + '\n')
		#fl.write('# Nd = ' + str(Nd) + '\n')
		#fl.write('# V0 = ' + str(V0) + '\n')
		fl.write('# Nint, Nbuf, Elow, Ggamma = ' + str(Nint) + ' ' + str(Nbuf)+ (' %12.8e'*2 % (EI,Ggamma)) + '\n')
		fl.write('# v = %12.8e \n' %(v))
		fl.write('# Re(k) Im(k)  Re(E)  Gamma_n \n')
		for i in range(len(kres)):
			kt = kres[i]; Et = E0[i]; Gamma = Gn[i]
			fl.write((' %12.8e '*4 + '\n') % (kt.real,kt.imag,Et,Gamma))

		#time = time.time()-start_time









